program wowpatcher;

{$mode objfpc}{$H+}

uses
  {$IFDEF UNIX}{$IFDEF UseCThreads}
  cthreads,
  {$ENDIF}{$ENDIF}
  Classes, SysUtils, CustApp
  { you can add units after this };

type

  { TPatcher }

  { TWoWPatcher }

  TWoWPatcher = class(TCustomApplication)
  private
    function FindSequence(Data, Sequence: array of Byte; Position: Integer = 0): Integer;
    procedure WriteHexData(Data: array of Byte; From, Count: Integer);
  protected
    procedure DoRun; override;
  public
    FileName: string;
    NewFileName: string;
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
    procedure WriteHelp; virtual;
    procedure Patch;
  end;

{ TWoWPatcher }

procedure TWoWPatcher.DoRun;
var
  ErrorMsg: String;
begin
  if HasOption('h', 'help') then begin
    WriteHelp;
    Terminate;
    Exit;
  end;
  if HasOption('s', 'source') then begin
    FileName := GetOptionValue('s', 'source');
  end;

  NewFileName := Copy(FileName, 1, Length(FileName) -
    Length(ExtractFileExt(FileName))) + '-2' + ExtractFileExt(FileName);
  Patch;

  Terminate;
end;

constructor TWoWPatcher.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  StopOnException := True;
  FileName := 'Wow.exe';
end;

destructor TWoWPatcher.Destroy;
begin
  inherited Destroy;
end;

procedure TWoWPatcher.WriteHelp;
begin
  Writeln('Usage: ', ExtractFileName(ExeName));
  WriteLn('    -s source  Select Wow.exe source file');
  WriteLn('    -h help    Show this help');
  WriteLn;
  WriteLn('Patched exe file cannot be used safely on offical servers.');
end;

function TWoWPatcher.FindSequence(Data, Sequence: array of Byte; Position: Integer = 0): Integer;
var
  I: Integer;
  II: Integer;
  Same: Boolean;
begin
  if Length(Sequence) = 0 then
    raise Exception.Create('Empty searched sequence');
  Result := -1;
  for I := Position to Length(Data) - Length(Sequence) do begin
    if Data[I] = Sequence[0] then begin
      Same := True;
      for II := 0 to High(Sequence) do begin
        if Data[I + II] <> Sequence[II] then begin
          Same := False;
          Break;
        end;
      end;
      if Same then begin
        Result := I;
        Break;
      end;
    end;
  end;
end;

procedure TWoWPatcher.WriteHexData(Data: array of Byte; From, Count: Integer);
var
  I: Integer;
begin
  for I := From to From + Count - 1 do
    Write(IntToHex(Data[I], 2) + ' ');
end;

procedure TWoWPatcher.Patch;
var
  OriginalFile: TFileStream;
  NewFile: TFileStream;
  Buffer: array of Byte;
  Position: Integer;
  I: Integer;
const
  StartSequence: array[0..9] of Byte = ($FF, $FF, $5B, $8B, $E5, $5D, $C3, $8D, $49, $00);
  StartSequence2: array[0..9] of Byte = ($E8, $71, $09, $FB, $FF, $5B, $8B, $E5, $5D, $C3);
begin
  OriginalFile := TFileStream.Create(FileName, fmOpenRead);
  NewFile := TFileStream.Create(NewFileName, fmCreate);
  SetLength(Buffer, OriginalFile.Size);
  OriginalFile.Read(Buffer[1], Length(Buffer));
  OriginalFile.Destroy;

  WriteLn;
  Write('Searching for first sequence ');
  WriteHexData(StartSequence, 0, Length(StartSequence));
  Write('...');
  Position := FindSequence(Buffer, StartSequence, 0);
  if Position <> -1 then begin
    WriteLn('found at ' + IntToHex(Position, 8));
    Inc(Position, Length(StartSequence));
    WriteLn('Changing data after sequence');
    WriteHexData(Buffer, Position, 16);
    for I := 0 to 11 do
      Buffer[Position + I] := Buffer[Position + 12 + (I mod 4)];
    WriteLn(' => ');
    WriteHexData(Buffer, Position, 16);
    WriteLn;
  end else WriteLn('not found');

  WriteLn;
  Write('Searching for second sequence ');
  WriteHexData(StartSequence2, 0, Length(StartSequence2));
  Write('...');
  Position := FindSequence(Buffer, StartSequence2);
  if Position <> -1 then begin
    WriteLn('found at ' + IntToHex(Position, 8));
    Inc(Position, Length(StartSequence2));
    WriteLn('Changing data after sequence');
    WriteHexData(Buffer, Position, 16);
    for I := 0 to 11 do
      Buffer[Position + I] := Buffer[Position + 12 + (I mod 4)];
    WriteLn(' => ');
    WriteHexData(Buffer, Position, 16);
    WriteLn;
  end else WriteLn('not found');

  NewFile.WriteBuffer(Buffer[0], Length(Buffer));
  NewFile.Destroy;
end;

var
  Application: TWoWPatcher;

{$R wowpatcher.res}

begin
  Application := TWoWPatcher.Create(nil);
  Application.Title := 'WoW patcher';
  Application.Run;
  Application.Free;
end.

